'''
@author: roj-idl71
'''
#USED IN jroplot_spectra.py
RTI_CODE = 0            #Range time intensity (RTI).
SPEC_CODE = 1           #Spectra (and Cross-spectra) information.
CROSS_CODE = 2          #Cross-Correlation information.
COH_CODE = 3            #Coherence map.
BASE_CODE = 4           #Base lines graphic.
ROW_CODE = 5            #Row Spectra.
TOTAL_CODE = 6          #Total Power.
DRIFT_CODE = 7          #Drifts graphics.
HEIGHT_CODE = 8         #Height profile.
PHASE_CODE = 9          #Signal Phase.

POWER_CODE = 16
NOISE_CODE = 17
BEACON_CODE = 18

#USED IN jroplot_parameters.py
WIND_CODE = 22
MSKYMAP_CODE = 23
MPHASE_CODE = 24

MOMENTS_CODE = 25
PARMS_CODE = 26
SPECFIT_CODE = 27
EWDRIFT_CODE = 28
